#include <cstdlib>
#include <ctime>
#include <GL/gl.h>
#include <GL/glu.h>

#include "wordz.h"

SDL_Surface* screen;
SDL_Surface** font;

GLuint textures[NUM_TEXTURES];

float ratio = 1.3333f;
float half_ratio = half_ratio;

char words[3][4];
int letter; // current letter
int skips = NUM_SKIPS;
int points[3];
int total_points;

enum MODE { GAME=1, HELP=2, SHOWPOINTS=3, VIEWSCORES=4, ENTERNAME=5, PROMPT=6 };
MODE mode = GAME;

char* Datapath(void)
{
	static char path[5000];
	char *env = getenv("WORDZ_DATA");

	if (env) {
		strncpy(path, env, sizeof(path)-1);
		if (path[strlen(path)-1] != '/')
			strncat(path, "/", sizeof(path)-1);
	}
	else {
#ifdef __linux__
		strncpy(path, DATAPATH, sizeof(path)-1);
#else
		strcpy(path, ".");
#endif
	}

	return path;
}

int GetPoints(int c)
{
	char* letters[5] = {
		"aeiou",
		"bdghlmnprstwy",
		"cfkv",
		"jxz",
		"q"
	};

	for (int i = 0; i < 5; i++) {
		for (int j = 0; letters[i][j]; j++)
			if (letters[i][j] == c)
				return i+1;
	}
	return 0;
}

void CalculatePoints(void)
{
	total_points = 0;

	for (int w = 0; w < 3; w++) {
		points[w] = 0;
		for (unsigned int i = 0; i < dictionary.size(); i++) {
			const char* s = dictionary[i].c_str();
			if (tolower(words[w][0]) == s[0] &&
					tolower(words[w][1]) == s[1] &&
					tolower(words[w][2]) == s[2]) {
				for (int l = 0; l < 3; l++) {
					points[w] += GetPoints(s[l]);
				}
				break;
			}
		}
		total_points += points[w];
	}
}

// Twice as many vowels
void GetNewLetter(void)
{
	int old = letter;
	letter = rand() % 26 + 'A';
	if (letter != 'A' && letter != 'E' && letter != 'I' && letter != 'O'
			&& letter != 'U')
		letter = rand() % 26 + 'A';
	if (old == letter)
		GetNewLetter();
}

// Returns true if word did not have 3 letters yet, otherwise false
bool PlaceLetter(int word)
{
	bool next = false;
	bool prev = false;

	for (int i = 0; i < 3; i++) {
		if (words[word][i] == 0) {
			if (word == 0 && i == 2) {
				if (words[1][0])
					return false;
				next = true;
			}
			else if (word == 1 && i == 2) {
				if (words[2][0])
					return false;
				next = true;
			}
			else if (word == 1 && i == 0)
				prev = true;
			else if (word == 2 && i == 0)
				prev = true;
			words[word][i] = letter;
			if (next) {
				word++;
				i = 0;
				words[word][i] = letter;
			}
			else if (prev) {
				word--;
				i = 2;
				words[word][i] = letter;
			}
			GetNewLetter();
			return true;
		}
	}

	return false;
}

void CheckGameOver(void)
{
	for (int w = 0; w < 3; w++)
		for (int i = 0; i < 3; i++)
			if (words[w][i] == 0)
				return;
	mode = SHOWPOINTS;
	CalculatePoints();
	//   in show points after key pressed or mouse, do high scores then
	//   clear words and start again
}

void ClearWords(void)
{
	for (int i = 0; i < 3; i++)
		for (int j = 0; j < 4; j++)
			words[i][j] = 0;
	skips = NUM_SKIPS;
}

void Quit(int return_code)
{
	SDL_Quit();
	exit(return_code);
}

bool LoadTextures(void)
{
	SDL_Surface** font;
	try {
		char filename[5000];
		snprintf(filename, 4999, "%s/font.bmp", Datapath());
		font = LoadFont(filename, FONT_W, FONT_H);
	}
	catch (std::bad_alloc) {
		fprintf(stderr, "bad alloc\n");
		return false;
	}
	if (!font) {
		fprintf(stderr, "font null\n");
		return false;
	}

	glGenTextures(NUM_TEXTURES, &textures[0]);
	for (int i = 0; i < NUM_TEXTURES; i++) {
		/* Typical Texture Generation Using Data From The Bitmap */
		glBindTexture(GL_TEXTURE_2D, textures[i]);

		/* Generate The Texture */
		//FIXME:
		glTexImage2D(GL_TEXTURE_2D, 0, 3, font[i]->w,
			font[i]->h, 0, GL_BGR,
			GL_UNSIGNED_BYTE, font[i]->pixels);

		/* Linear Filtering */
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
	
	DeleteFont(font);

	return true;
}

bool InitializeGL(void)
{
	if (!LoadTextures())
		return false;

	/* Enable Texture Mapping */
	glEnable(GL_TEXTURE_2D);

	/* Enable smooth shading */
	glShadeModel(GL_SMOOTH);

	/* Set the background black */
	glClearColor(0.0f, 0.0f, 0.0f, 0.5f);

	/* Depth buffer setup */
	glClearDepth(1.0f);

	/* Enables Depth Testing */
	glEnable(GL_DEPTH_TEST);

	/* The Type Of Depth Test To Do */
	glDepthFunc(GL_LEQUAL);

	/* Really Nice Perspective Calculations */
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	return true;
}

void ResizeWindow(int width, int height)
{
	/* Protect against a divide by zero */
	if ( height == 0 )
		height = 1;

	ratio = (GLfloat)width / (GLfloat)height;
	half_ratio = ratio / 2.0f;
	char_w = ratio / 33.0f;
	printed_char_w = char_w / 64.0f * (float)(64-char_padding_h);

	/* Setup our viewport. */
	glViewport(0, 0, (GLint)width, (GLint)height);

	/*
	* change to the projection matrix and set
	* our viewing volume.
	*/
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	/* Set our perspective */
	gluPerspective(45.0f, ratio, 0.1f, 100.0f);

	/* Make sure we're chaning the model view and not the projection */
	glMatrixMode(GL_MODELVIEW);

	/* Reset The View */
	glLoadIdentity();

	SDL_WM_SetCaption("WordZ", 0);
	SetIcon();
}

void HandleKeyPress(SDL_keysym *keysym)
{
	if (keysym->sym == SDLK_F1) {
		SDL_WM_ToggleFullScreen(screen);
		return;
	}

	if (mode == GAME) {
		switch (keysym->sym) {
			case SDLK_ESCAPE:
				//FIXME: prompt are you sure?
				Quit(0);
				break;
			case SDLK_1:
				if (PlaceLetter(0))
					CheckGameOver();
				else
					;// error sound
				break;
			case SDLK_2:
				if (PlaceLetter(1))
					CheckGameOver();
				else
					;//error sound
				break;
			case SDLK_3:
				if (PlaceLetter(2))
					CheckGameOver();
				else
					;//error sound
				break;
			case SDLK_s:
				if (skips) {
					GetNewLetter();
					skips--;
				}
				else {
					// error sound
				}
				break;
			case SDLK_h:
				mode = HELP;
				break;
			default:
				break;
		}
	}
	else if (mode == HELP)
		mode = GAME;
	else if (mode == SHOWPOINTS) {
		ClearWords();
		skips = NUM_SKIPS;
		mode = GAME; //FIXME: mode=ENTERNAME if high score
	}
}

void DrawWords(void)
{
	struct {
		int x;
		int y;
	} loc[3][3] = {
		{
			{ 76, 91 },
			{ 151, 91 },
			{ 0, 0 }
		},
		{
			{ 226, 91 },
			{ 226, 166 },
			{ 0, 0 },
		},
		{
			{ 226, 241 },
			{ 301, 241 },
			{ 376, 241 }
		}
	};

	int x, y;

	for (int w = 0; w < 3; w++) {
		for (int i = 0; i < 3; i++) {
			if (i == 2 && w < 2)
				continue;
			if (words[w][i]) {
				x = loc[w][i].x;
				y = loc[w][i].y;
				GLPrintf(x, y, "%c", words[w][i]);
			}
		}
	}
}

void Draw(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
	glTranslatef(0.0f, 0.0f, -2.2f);
    
	glBegin(GL_QUADS);
		glColor3f(1.0f, 0.58f, 0.0f); glVertex3f( -half_ratio, -0.5f, 1.0f );
		glColor3f(1.0f, 0.98f, 0.4f); glVertex3f(  half_ratio, -0.5f, 1.0f );
		glColor3f(1.0f, 0.58f, 0.0f); glVertex3f(  half_ratio,  0.5f, 1.0f );
		glColor3f(1.0f, 0.58f, 0.0f); glVertex3f( -half_ratio,  0.5f, 1.0f );
	glEnd();

	if (mode == GAME) {
		GLPrintf(500, 360, "Letter    Points");
		GLPrintf(500, 390, "AEIOU..........1");
		GLPrintf(500, 420, "BDGHLMNPRSTWY..2");
		GLPrintf(500, 450, "CFKX...........3");
		GLPrintf(500, 480, "JXZ............4");
		GLPrintf(500, 510, "Q..............5");
		GLPrintf(500, 540, "Press H for help");

		Uint32 black = SDL_MapRGB(screen->format, 0, 0, 0);
		Uint32 orange = SDL_MapRGB(screen->format, 255, 147, 0);

		DrawRect(50, 70, 80, 80, black, orange);
		DrawRect(125, 70, 80, 80, black, orange);
		DrawRect(200, 70, 80, 80, black, orange);
		DrawRect(200, 145, 80, 80, black, orange);
		DrawRect(200, 220, 80, 80, black, orange);
		DrawRect(275, 220, 80, 80, black, orange);
		DrawRect(350, 220, 80, 80, black, orange);

		GLPrintf(78, 30, "1");
		GLPrintf(228, 30, "2");
		GLPrintf(170, 245, "3");

		DrawWords();

		GLPrintf(535, 70, "Your letter:");
		GLPrintf(635, 120, "%c", letter);
		GLPrintf(590, 190, "Skips:");
		GLPrintf(620, 240, "%d", skips);
	}
	else if (mode == HELP) {
		GLPrintf(100, 100, "Make words to score points!");
		GLPrintf(100, 130, "To place a letter, press the");
		GLPrintf(100, 160, "number of the word you want");
		GLPrintf(100, 190, "to place it in:");

		GLPrintf(100, 250, "1) The first word");
		GLPrintf(100, 280, "2) The second word");
		GLPrintf(100, 310, "3) The third word");
		GLPrintf(100, 340, "S) Skip this letter");

		GLPrintf(100, 400, "You can skip up to 10 letters");
		GLPrintf(100, 430, "per game.");
		GLPrintf(300, 500, "Press any key to continue..");
	}
	else if (mode == SHOWPOINTS) {
		GLPrintf(100, 100, "You scored %d points!", total_points);
		GLPrintf(100, 150, "%s: %d points", words[0], points[0]);
		GLPrintf(100, 180, "%s: %d points", words[1], points[1]);
		GLPrintf(100, 210, "%s: %d points", words[2], points[2]);
		GLPrintf(300, 500, "Press any key to continue..");
	}


	SDL_GL_SwapBuffers( );

static GLuint Frames = 0;
static GLuint To = 0;

Frames++;
GLint t = SDL_GetTicks();
if (t - To >= 5000) {
    GLfloat seconds = (t - To) / 1000.0;
    GLfloat fps = Frames / seconds;
    To = t;
    Frames = 0;
}
}

#ifdef __linux__
int main(int argc, char** argv)
#else
int WINAPI WinMain(      
    HINSTANCE hInstance,
    HINSTANCE hPrevInstance,
    LPSTR lpCmdLine,
    int nCmdShow)
#endif
{
	int video_flags;
	const SDL_VideoInfo* video_info;
	bool done = false;
	bool is_active = true;
	SDL_Event event;

	ReadDictionary();

	srand(time(NULL));

	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "Video intialization failed: %s\n",
				SDL_GetError());
		Quit(1);
	}

	video_info = SDL_GetVideoInfo();

	if (!video_info) {
		fprintf(stderr, "Video query failed: %s\n",
				SDL_GetError());
		Quit(1);
	}

	video_flags = SDL_OPENGL;
	video_flags |= SDL_GL_DOUBLEBUFFER;
	video_flags |= SDL_HWPALETTE;
	video_flags |= SDL_RESIZABLE;

	if (video_info->hw_available) {
		printf("hw\n");
		video_flags |= SDL_HWSURFACE;
		cfg.has_hw_surfaces = true;
	}
	else {
		video_flags |= SDL_SWSURFACE;
		cfg.has_hw_surfaces = false;
	}

	if (video_info->blit_hw)
		video_flags |= SDL_HWACCEL;

	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

	cfg.bpp = SCREEN_BPP;

	screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, cfg.bpp,
			video_flags);

	if (!screen) {
		fprintf(stderr, "Setting video mode failed: %s\n",
				SDL_GetError());
		Quit(1);
	}

	if (!InitializeGL()) {
		fprintf(stderr, "Error loading textures\n");
		Quit(1);
	}

	ResizeWindow(SCREEN_WIDTH, SCREEN_HEIGHT);

	ClearWords();
	GetNewLetter();
    
	while (!done) {
	/* handle the events in the queue */
		while (SDL_PollEvent(&event)) {
			switch(event.type) {
				case SDL_ACTIVEEVENT:
					/* Something's happend with our focus
					* If we lost focus or we are iconified, we
					* shouldn't draw the screen
					*/
					if ( event.active.gain == 0 )
						is_active = false;
					else
						is_active = true;
					break;			    
				case SDL_VIDEORESIZE:
					/* handle resize event */
					screen = SDL_SetVideoMode(
							event.resize.w,
							event.resize.h,
							16, video_flags);
					if (!screen) {
						fprintf(stderr, "Could not get a surface after resize: %s\n", SDL_GetError());
						Quit(1);
					}
					ResizeWindow(event.resize.w, event.resize.h);
					break;
				case SDL_KEYDOWN:
					/* handle key presses */
					HandleKeyPress(&event.key.keysym);
					break;
				case SDL_QUIT:
					/* handle quit requests */
					done = true;
					break;
				default:
					break;
			}
		}

		/* draw the scene */
		if (is_active)
			Draw();
	}

	Quit(0);

	return 0;
}
